# -*- coding: utf-8 -*-
"""
Created on Sun May  2 16:58:04 2021
author: Urs Leisinger
email: urs.leisinger@bluewin.ch

Vorlage:
https://jakevdp.github.io/blog/2012/08/18/matplotlib-animation-tutorial/
Animation of Elastic collisions with Gravity
author: Jake Vanderplas
email: vanderplas@astro.washington.edu
website: http://jakevdp.github.com
license: BSD
Please feel free to use and modify this, but keep the above information. Thanks!

events:
    
https://matplotlib.org/stable/users/event_handling.htm


Konstante Energien
Die Reflexion am gekrümmten Boden macht immer wieder Korrekturen nötig, welche die GEsamtenergie verändern
Daher muss die Energie künstlich konstant gehalten werden. Damit die Teilchen am Anfang fallen können, wird die GEsamtenergie konstant gehalten. 
Will man hingegen die Niveaus verändern, gibt das ein Problem: Verschiebt man ein Niveau nach oben, sind die Teilchen höher und ihre potentielle Energie höher
würde man die Gesamtenergie konstant halten, nähme die "Temperatur" ab. Im vorliegenden Programm wird das so gelöst, dass bei Niveauveränderungen auf einen anderen Modus umgestellt wird, bei dem die kinetische Energie konstant gehalten wird. 
Beim Hinzufügen von Teilchen ergibt sich ein ähnliches Problem. Dort wird die Soll-GEsamtenergie proportional zur Teilchenzahl skaliert.


# ------------------------------------------------------------------------------
# Überprüfung der Beziehung K = exp(deltaG/RT)
- nicht zu viele Teilchen, nicht zu grosse Teilchen (das totale Teilchenvomlumen muss vernachlässigbar bleiben)
- da nur 2 Dimensionen = 2 Freiheitsgrade: <Ekin> = 2 * 1/2 kT 
- <Ekin> pro Teilchen wird alle 500 Berechnungen ausgegeben
- deltaG entspricht dem Unterschied der potentiellen Energie, wird alle 500 Berechnungen ausgegeben
- Weil die Teilchen Kinetische Energie verlieren, wenn sie auf das obere Niveau wechseln, sind die Teilchen auf dem oberen Niveau
stets langsamer, "kälter". mit der Taste "kick" wird Ekin neu verteilt, also ein Temperaturgleichgewicht wieder hergestellt
Damit die Beziehung stimmt, muss man alle paar Sekunden "kick" drücken.


Bilder in Film umwandeln mit ffmpeg
ffmpeg -start_number 0 -i animation_%04d.png -filter:v fps=30  -vf "pad=ceil(iw/2)*2:ceil(ih/2)*2" -pix_fmt yuv420p output.mp4


Linienfarben:
import matplotlib as mpl
mpl.rcParams["axes.prop_cycle"]
Out[14]: cycler('color', ['#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd', '#8c564b', '#e377c2', '#7f7f7f', '#bcbd22', '#17becf'])

"""
import numpy as np
from scipy.spatial.distance import pdist, squareform

import matplotlib.pyplot as plt
import scipy.integrate as integrate
import matplotlib.animation as animation
from matplotlib.widgets import Button, TextBox
import matplotlib.gridspec as gridspec
import matplotlib.patches as patches

mehrere_niveaus = False  # wenn x_uz < 0.5, werden mehrere Niveaus in immer gleichen Abständen gezeichnet (oder nur 1 langes Produkte-Niveau bei False)
blitten = False         # muss False sein, damit man Bilder speichern kann

bereiche_separat = True  #Energie konstanthalten: Edukte und Produkte separat oder alle gemeinsam

x_uz = 0.5 #x-Position des Übergangszustandes (von 0 bis 1)

#Bereich, in dem die Teilchen erzeugt werden:
x_teilchen_min = 0.05
x_teilchen_max = 0.45

#Teilchen
n_teilchen = 60             #Anzahl 50
r_teilchen = 0.01           #Durchmesser 0.005

#Energien
E1 = 0.1                    # Niveau 1
EA = 0.45                    # Aktivierungsenergie
E2 = 0.5                    # Niveau 2

Ekin_nachjustieren = False   # Wenn Ekin konstant gehalten wird: die Ekin nach einigen Zyklen wird zum Sollwert
#------------------------------------------------------------
# in dieser Klasse wird die Form des Bodens festgelegt
class Boden:
    def __init__(self, e1=1, e2=0.5, ea1 = 1, x_uz = 0.5, mehrere_niveaus=False):
        """
        Parameters
        ----------
        e1 : TYPE, optional
            Energie der Edukte. The default is 1. 
        e2 : TYPE, optional
            Energie der Produkte. The default is 0.5.
        ea1 : TYPE, optional
            Aktivierungsenergie Hinreaktion. The default is 1.
        x_uz : TYPE, optional
            Position des Übergangszustandes auf der x-Achse. The default is 0.

        Returns
        -------
        None.
        
        Baut eine einigermassen typisch aussehenden Energieverlauf für eine elementare Reaktion zusammen, 
        die auf die übergebenen Paramter passt. self.ea2 ist die Aktivierungsenergie der Rückreaktion

        """
        self.e1 = e1
        self.e2 = e2
        self.ea1 = ea1                                       # Aktivierungsenergie der Rückreaktion
        self.x_uz = x_uz
        self.n_points = 10000
        self.x = np.linspace(0,1,self.n_points)
        self.bodenlinie, = ax.plot([],[],'k-', lw = 2)
        self.mehrere_niveaus = mehrere_niveaus
        self.boden_zeichnen()
        
    def boden_zeichnen(self):
        if self.mehrere_niveaus == False: 
            self.y = np.zeros(self.n_points)
            self.ea2 = self.e1 + self.ea1 - self.e2 
            self.sigma1 = 0.1 * np.sqrt(self.ea1)                           # Peakbreite für den Kurvenverlauf links vom Übergangszustand
            self.sigma2 = 0.1 * np.sqrt(self.ea2)                           # Peakbreite für den Kurvenverlauf rechts vom Übergangszustand
            y1a = np.exp(-(((self.x-self.x_uz)/(self.sigma1))**2))               # Kurvenverlauf links vom Übergangszustand
            y1b = np.exp(-(((self.x-self.x_uz)/(self.sigma2))**2))               # Kurvenverlauf rechts vom Übergangszustand
            self.y = self.ea1 * y1a + self.e1                               # Boden links vom Übergangszustand (rechter Teil nach derselben Formel, wird in der nächsten Zeile korrigiert)
            self.y[self.x > self.x_uz]= self.ea2 * y1b[self.x > self.x_uz] + self.e2  # Boden rechs vom Übergangszustand
        else: 
            x_uz_n = self.x_uz
            self.y = np.zeros(self.n_points)
            self.ea2 = self.e1 + self.ea1 - self.e2
            delta_e = self.e2-self.e1
            e1 = self.e1
            e2 = self.e2
            ea1 = self.ea1
            icount = 0
            while x_uz_n < 1:           
                ea2 = e1 + ea1 - e2 
                self.sigma1 = 0.1 * np.sqrt(ea1)                           # Peakbreite für den Kurvenverlauf links vom Übergangszustand
                self.sigma2 = 0.1 * np.sqrt(ea2)                           # Peakbreite für den Kurvenverlauf rechts vom Übergangszustand
                y1a = np.exp(-(((self.x-x_uz_n)/(self.sigma1))**2))               # Kurvenverlauf links vom Übergangszustand
                y1b = np.exp(-(((self.x-x_uz_n)/(self.sigma2))**2))               # Kurvenverlauf rechts vom Übergangszustand
                if icount == 0:
                    self.y = ea1 * y1a + e1                               # Boden links vom Übergangszustand (rechter Teil nach derselben Formel, wird in der nächsten Zeile korrigiert)
                else:
                    self.y[self.x > (x_uz_n - self.x_uz/2)] = ea1 * y1a[self.x > (x_uz_n - self.x_uz/2)] + e1
                self.y[self.x > x_uz_n]= ea2 * y1b[self.x > x_uz_n] + e2  # Boden rechs vom Übergangszustand
                e1 = e2
                e2 = e2 + delta_e
                x_uz_n += self.x_uz
                icount += 1
                print("Niveau gezeichnet:", icount)
        

class ParticleBox:
    """Orbits class
    
    init_state is an [N x 4] array, where N is the number of particles:
       [[x1, y1, vx1, vy1],
        [x2, y2, vx2, vy2],
        ...               ]

    bounds is the size of the box: [xmin, xmax, ymin, ymax]
    """
    def __init__(self,
                 init_state = [[1, 0, 0, -1],
                               [-0.5, 0.5, 0.5, 0.5],
                               [-0.5, -0.5, -0.5, 0.5]],
                 bounds = [0, 1, 0, 1],
                 radius = 0.05,
                 M = 0.05,
                 G = 9.8, 
                 konstant = "ekin",
                 bereiche_separat = False): #etot, ekin
        self.init_state = np.asarray(init_state, dtype=float)
        self.masse = M
        self.M = self.masse * np.ones(self.init_state.shape[0])
        self.radius = radius
        self.state = self.init_state.copy()
        self.time_elapsed = 0
        self.bounds = bounds
        self.G = G
        self.Q = [[]] #Reaktionsquotient
        self.t_Q = []
        self.Q_mittel_laenge = 1
        self.i_Q_mittel = [0]
        self.Q_mittel_vals = [[]]
        self.bereiche_separat = bereiche_separat
        
        self.zwischenschritte = 1
        
        self.alle_anheben()

        self.i_count = 0
        # Totale Energie wird berechnet und gespeichert, damit Abdriften der GEsamtenergie anschliessend korrigiert werden kann
        self.Ekin_tot_0 = np.sum(0.5 * self.M * np.sum(self.state[:,2:]**2, axis=1))
        self.Epot_tot_0 = np.sum(self.state[:,1]*self.M*self.G)
        self.Etot_tot_0 = self.Ekin_tot_0 + self.Epot_tot_0

        self.Ekin_sum =0 
        self.Ekin_zeitschritte =0 
        
        self.konstant = konstant #welche Energie konstant gehalten werden soll: 'etot' oder 'ekin'

        # Anzahl Niveaus und Anzahl Q
        if epot_boden.mehrere_niveaus == False:
                self.n_niveaus = 2
        else:
            self.n_niveaus = int(np.ceil(1/epot_boden.x_uz))
        self.n_Qs = self.n_niveaus - 1 
        self.Q_initialisieren()
        



    def Q_initialisieren(self):
        self.Q_mittel_vals = []
        self.i_Q_mittel = []
        self.Q = []
        #ax2.lines = []
        self.verlaeufe = []

        

        self.farben = ['#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd', '#8c564b', '#e377c2', '#7f7f7f', '#bcbd22', '#17becf']
           
        for iq in range(self.n_Qs):
            self.Q_mittel_vals.append([])
            self.i_Q_mittel.append(0)
            self.Q.append([])
            if self.n_Qs > 1:
                label = "Q"+str(iq+1)
            else: 
                label = "Q"
            verlauf, = ax2.plot([], [], label=label)
            self.verlaeufe.append(verlauf)
        
        self.legende = ax2.legend()        

    def step(self, dt):
        """step once by dt seconds"""
        self.time_elapsed += dt
        
        # update positions
        self.state[:, :2] += dt * self.state[:, 2:]

        # find pairs of particles undergoing a collision
        D = squareform(pdist(self.state[:, :2]))
        ind1, ind2 = np.where(D < 2 * self.radius)    # wenn alle Teilchen weit genug voneinander weg sind, wird hier jedes Teilchen mit sich selber ausgegeben
        unique = (ind1 < ind2) # Sonst kolliidert jedes Teilchen mit sich selber (ind1 == ind2) und jede Kollision wird doppelt gezählt. 
        ind1 = ind1[unique]    
        ind2 = ind2[unique]
            
        # update velocities of colliding pairs
        for i1, i2 in zip(ind1, ind2):
            # mass
            m1 = self.M[i1]
            m2 = self.M[i2]

            # location vector
            r1 = self.state[i1, :2]
            r2 = self.state[i2, :2]

            # velocity vector
            v1 = self.state[i1, 2:]
            v2 = self.state[i2, 2:]

            # relative location & velocity vectors
            r_rel = r1 - r2
            v_rel = v1 - v2

            # momentum vector of the center of mass
            v_cm = (m1 * v1 + m2 * v2) / (m1 + m2)

            # collisions of spheres reflect v_rel over r_rel
            r_rel_ev = r_rel / (np.dot(r_rel, r_rel))**0.5           # Einheitsvektor in Richtung des Verbindungsvektors
            vr_rel = np.dot(v_rel, r_rel_ev)
            v_rel = 2 * r_rel_ev * vr_rel - v_rel

            # assign new velocities #falsch herum im Original
            self.state[i1, 2:] = v_cm - v_rel * m2 / (m1 + m2)
            self.state[i2, 2:] = v_cm + v_rel * m1 / (m1 + m2) 
            
            #Teilchen auseinanderrücken: entlang der neuen Geschwindigkeiten einen zeitschritt ausführen
            self.state[i1, :2] += dt * self.state[i1, 2:]
            self.state[i2, :2] += dt * self.state[i2, 2:]
            #falls Teilchen (von Beginn weg) ineinander sind:
            d_paar  = np.linalg.norm(self.state[i1,:2] - self.state[i1,:2]) 
            if d_paar < 2 * self.radius:
                r_cm = (r1 + r2)/2
                self.state[i1, :2] = r_cm + (self.radius*2 - d_paar)/2 * r_rel_ev
                self.state[i2, :2] = r_cm - (self.radius*2 - d_paar)/2 * r_rel_ev
                

        # check for crossing boundary
        crossed_x1 = (self.state[:, 0] < self.bounds[0] + self.radius)
        crossed_x2 = (self.state[:, 0] > self.bounds[1] - self.radius)
        crossed_y1 = (self.state[:, 1] < self.bounds[2] + self.radius)
        #crossed_y2 = (self.state[:, 1] > self.bounds[3] - self.radius)

        self.state[crossed_x1, 0] = self.bounds[0] + self.radius
        self.state[crossed_x2, 0] = self.bounds[1] - self.radius

        #self.state[crossed_y1, 1] = self.bounds[2] + self.radius          #untere Begrenzung überschritten, buggt, wenn zugleich der Boden bei Null ist
        #self.state[crossed_y2, 1] = self.bounds[3] - self.radius          #Obere Begrenzung überschritten

        self.state[crossed_x1 | crossed_x2, 2] *= -1
        #self.state[crossed_y1 | crossed_y2, 3] *= -1                    # obere oder untere Begrenzung überschritten
        

        # Reflexion am Boden
        # Es werden alle Abstände zwischen dem Teilchen und jedem Bodenpunkt berechnet
        # ist der kleinste Abstand zu klein, so wird der entsprechende Punkt gesucht und das Teilchen an diesem Punkt reflektiert
        for i in range(len(self.state)):

            x,y = self.state[i,:2]
            distanz = ((x-epot_boden.x)**2+(y-epot_boden.y)**2)**0.5
            minimale_distanz = np.amin(distanz)

            if minimale_distanz < self.radius:
                i_minimum = np.where(distanz == minimale_distanz)[0][0]
                b_punkt = np.array([epot_boden.x[i_minimum],epot_boden.y[i_minimum]])  #Berührungspunkt
                r_verbindung = self.state[i,:2] - b_punkt
                # Reflexion an der Senkrechten zur Verbingunslinie zwischen nächstem Bodenpunkt und Teilchen r_verbindung
                # wir projizieren die Geschwindigkeit auf den normierten Verbindungsvektor. 
                # zieht man das Resultat zwei mal von der Geschwindigkeit ab, erhält man die neue Geschwindigkeit 
                richtung = r_verbindung/np.linalg.norm(r_verbindung)
                self.state[i,2:] -= richtung * 2 * np.dot(self.state[i,2:], r_verbindung)/np.linalg.norm(r_verbindung)
                self.state[i,:2] += richtung * (1.0 * self.radius-minimale_distanz)               #damit Teilchen nicht zu tief geraten und dann unter dem Boden hängen bleiben
            
            # Teilche, die zu schnell sind, durchschlagen den Boden, ohne ihn zu berühren
            # Diese Teilchen werden hier wieder angehoben. Allerdings erhöht dies die Gesamtenergie (Korrektur siehe unten)
            self.anheben(i)
            
        # Anzahl Niveaus und Anzahl Q
        if epot_boden.mehrere_niveaus == False:
                self.n_niveaus = 2
        else:
            self.n_niveaus = int(np.ceil(1/epot_boden.x_uz))
        self.n_Qs = self.n_niveaus - 1 
        
           
        # Teilchen
        self.n_teilchen = self.state.shape[0]
        #self.n_edukte = len(np.where(self.state[:,0] < epot_boden.x_uz)[0])
        #self.n_produkte = len(np.where(self.state[:,0] >= epot_boden.x_uz)[0])

        self.teilchen_liste = []
        for i in range(self.n_Qs):
            #Bei nur zwei Niveaus: Edukte = self.teilchen_liste[0], Produkte = teilchen.iste[1]
            xlinks = 0
            x_uz_n = epot_boden.x_uz
            xrechts = x_uz_n
            zaehler = 0
            while xlinks < 1:
                teilchen_links = len(np.where(self.state[:,0] < xlinks)[0])
                teilchen_rechts = len(np.where(self.state[:,0] < xrechts)[0])
                self.teilchen_liste.append(teilchen_rechts - teilchen_links)
                xlinks = xrechts
                xrechts += epot_boden.x_uz
                if epot_boden.mehrere_niveaus == False:
                    xrechts = 1

        #durch das Anheben werden die Teilchen immer schneller. Wird hier korrigiert
        #Option: jeder Bereich wird unabhängig vom anderen korrigiert (nur ekin).
        self.Ekin_tot = np.sum(0.5 * self.M * np.sum(self.state[:,2:]**2, axis=1))
        self.Epot_tot = np.sum(self.state[:,1]*self.M*self.G)
        self.Etot_tot = self.Ekin_tot + self.Epot_tot 
        self.Ekin_sum += self.Ekin_tot
        self.Ekin_zeitschritte += 1
        self.Ekin_avg_0 = self.Ekin_tot_0/self.n_teilchen

        # Listen vorbereiten  
        if len(self.Q_mittel_vals) != self.n_Qs:
            self.Q_initialisieren()





        # bei zu grosser Gesamtenergie wird die Geschwindigkeit aller Teilchen in jeder Richtung um 1% gesenkt
        if self.konstant == 'etot':
            if self.Etot_tot > self.Etot_tot_0:
                self.state[:,3] *= 0.999   #Dämpfung nur in y-Richtung
            if self.Etot_tot < self.Etot_tot_0:
                self.state[:,3] *= 1.001
                
        elif self.konstant == 'ekin':
            if self.bereiche_separat == False:
                # Bereiche vor und nach jedem Übergangszustand separat dämpfen oder beschleunigen
                if self.Ekin_tot > self.Ekin_tot_0:
                    self.state[:,3] *= 0.999   #Dämpfung nur in y-Richtung
                if self.Ekin_tot < self.Ekin_tot_0:
                    self.state[:,3] *= 1.001    
            else:                  
                for i in range(self.n_niveaus):
                    x1 = epot_boden.x_uz * i
                    x2 = epot_boden.x_uz * (i+1)
                    n_indices = np.where(np.logical_and(self.state[:,0] >= x1 , self.state[:,0] < x2))
                    n_dieses_niveau = len(n_indices[0])
                    Ekin_tot_niveau = np.sum(0.5 * self.M[n_indices[0]] * np.sum(self.state[n_indices[0],2:]**2, axis=1))
                    #print("ekin vergleich", Ekin_tot_niveau , self.Ekin_avg_0 * n_part, self.Ekin_avg_0 * self.n_teilchen, self.n_teilchen)
                    if Ekin_tot_niveau > self.Ekin_avg_0 * n_dieses_niveau:
                        self.state[n_indices[0],3] *= 0.999   #Dämpfung nur in y-Richtung
                    else:
                        self.state[n_indices[0],3] *= 1.001 




            
        # add gravity
        self.state[:, 3] -= self.G * dt
        
        # Teilchengrösse bestimmen für den Fall dass Fenstergrösse verändert wird
        #ms = int(fig.dpi * 2 * box.size * fig.get_figwidth() / np.diff(ax.get_xbound())[0])
        # wie genau der Faktor 144 zustande kommt, ist unklar, aber er passt auf alle Fenstervarianten
        bbox = ax.get_window_extent().transformed(fig.dpi_scale_trans.inverted())
        self.markersize = int(144* self.radius * bbox.width / np.diff(ax.get_xbound())[0])


        # ----------------------------------------------
        # Darstellung des Verlaufs
        
        Q_len_max = 500
        for iq in range(self.n_Qs):
            if (self.i_Q_mittel[iq] < self.Q_mittel_laenge) and (self.teilchen_liste[iq] > 0):  # sonst wird Wert nan
                self.Q_mittel_vals[iq].append(self.teilchen_liste[iq+1]/self.teilchen_liste[iq])
                self.i_Q_mittel[iq] += 1
            else:
                self.Q_mittel_vals[iq].append(1E6)                                  #statt unendlich wird ein riesiger Wert angefügt
                self.i_Q_mittel[iq] += 1
            
            if (self.i_Q_mittel[iq] >= self.Q_mittel_laenge) and (len(self.Q_mittel_vals[iq]) > 0): 
                self.i_Q_mittel[iq] = 0
                self.Q[iq].append(np.average(self.Q_mittel_vals[iq]))
                self.Q_mittel_vals[iq] = []
        

        
        
            #hier wird der Array jeweils halbiert und der Durchschnitt zweier aufeinanderfolgender Werte genommen
        if len(self.Q[0])<200:
            t_Q_max = 2
        else: 
            t_Q_max = 1
        if len(self.Q[0]) >= Q_len_max:
            self.Q_mittel_laenge *= 2
            
                #print("Edukte: {}, Produkte: {}, durchschnittliches Q: {}".format(self.n_edukte, self.n_produkte, np.average(self.Q)))

            for iq in range(self.n_Qs):
                if len(self.Q[iq]) >= Q_len_max:
                    self.Q[iq] = list(np.average(np.array(self.Q[iq]).reshape(-1,2), axis=1))
        self.t_Q = np.linspace(0, t_Q_max, 510)          

                






        """
        
        if epot_boden.mehrere_niveaus == True:
            if len(self.Q_mehrere_niveaus) != len(self.teilchen_liste)-1:
                self.Q_mehrere_niveaus = []
                self.i_count = 0
                for i in range(len(self.teilchen_liste)-1):
                    self.Q_mehrere_niveaus.append(self.teilchen_liste[i+1]/self.teilchen_liste[i])
            else: 
                 for i in range(len(self.teilchen_liste)-1):
                     try:
                         self.Q_mehrere_niveaus[i] = (self.Q_mehrere_niveaus[i] * self.i_count +  self.teilchen_liste[i+1]/self.teilchen_liste[i])/(self.i_count + 1)
                     except:
                         print("Fehler: Qs, i_count, teilchen_liste,  Verhältnis",self.Q_mehrere_niveaus, self.i_count, self.teilchen_liste, self.teilchen_liste[i+1]/self.teilchen_liste[i] )
            self.Q_mehrere_nivs_liste.append(self.Q_mehrere_niveaus)
            if self.i_count % 50 == 0:
                text = str(self.i_count) + ", mittleres "
                for i in range(len(self.Q_mehrere_niveaus)):
                    text += " Q"+str(i) + " = {:.05g}".format(self.Q_mehrere_niveaus[i])+" "
                text += " box.Ekin_tot_0 = " + str(box.Ekin_tot_0) 
                print(text)
        """
        # mittlere Q in Konsole drucken
        if self.i_count % 100 == 0:
            Qarray = np.array(self.Q)
            if box.konstant == "ekin":
                konstante_E = "E_kin"
            else: 
                konstante_E = "E_tot"
            text = "i = {}, konstant: {}, mittlere".format(self.i_count, konstante_E)
            for i in range(len(Qarray)):
                text += " Q"+str(i+1) + " = {:.05g},".format(np.average(np.array(Qarray)[i]))+" "
            text += "Ekin_tot_0 ={:.04g}".format(box.Ekin_tot_0) 
            print(text)
            
        self.i_count += 1        
        
        
        
    # Teilchen über den Boden anheben
    def anheben(self,i, zufall = 0):
        index = np.where(epot_boden.x >= self.state[i,0] - self.radius)                      # alle Punkte des  Bodens beim betrachteten Teilchen oder rechts davon
        if epot_boden.y[index[0][0]] > self.state[i,1]:                       # epot_boden.y[index[0][0]] ist das erste Bodenelement von der obigen Auswahl - also das Bodenelement genau unter dem Teilchen 
            self.state[i,1] += epot_boden.y[index[0][0]]-self.state[i,1] + self.radius + np.random.random(1)*zufall #zufall bewirkt, dass die Punkte zufällig noch etwas höher angehoben werden als nur gerade über den Boden    

    def alle_anheben(self):
        for i in range(len(self.state)):
            self.anheben(i, zufall = 0.1)         

#------------------------------------------------------------
# Figur und Animation vorbereiten

fig = plt.figure(figsize = (12,6)) # constrained_layout=True)
gs = gridspec.GridSpec(10, 16, wspace=0.2, left = 0.03, right= 0.99)

ax = fig.add_subplot(gs[:,:8])
ax2 = fig.add_subplot(gs[:7, 8:])

"""
blit führt (auf Mac) unmittelbar ausserhalb der Grafiken zu Löschungen. 
Wenn man wspace vergrössert, vergrössern sich auch die gelöschten Bereiche.
Bisher einzige Lösung: Schrift klein machen und mittels pad etwas vom Rand der Grafik wegrücken.   
"""
ax2.get_yaxis().set_tick_params(pad=8)
ax2.tick_params(axis='y', labelsize=6)


#------------------------------------------------------------
# Anfangszustand einrichten

i_bild = 0


epot_boden = Boden(e1=E1, e2 = E2, ea1 = EA, x_uz = x_uz, mehrere_niveaus = mehrere_niveaus)

np.random.seed(0)

def box_initialisieren(n_teilchen = n_teilchen):
    init_state = -0.5 + np.random.random((n_teilchen, 4)) #alle Werte liegen zwischen -0.5 und 0.5
    init_state[:, 1] += 0.5
    init_state[:, 1] *= 1.0             #Startpositioinen in y-Richtung
    init_state[:, 0] += 0.5           # x-Werte zwischen 0 und 1
    init_state[:, 0] *= x_teilchen_max-x_teilchen_min #Startpositionen in x-Richtung. Man könnte die Teilchen auch über den ganzen Bereich verteilen oder auf die Produktseite verschieben
    init_state[:, 0] += x_teilchen_min
    init_state[:, 2:] *= 1             #Startgeschwindigkeiten justieren

    return ParticleBox(init_state, radius=r_teilchen, konstant = "ekin", bereiche_separat=bereiche_separat)

box = box_initialisieren()

## fixes Ekin_tot hier einstellen
box.Ekin_tot_0 = 2.

legend_step = 1


dt = 0.05 / 30 # 30fps


# particles ist eine Linie, bei der jede Ecke ein Partikel ist, bei der nur Marken, aber keine Linien gezeichnet werden
particles, = ax.plot([], [], marker = 'o', color = 'b', linestyle = '')              #markersize wird in der Animation festgelegt (nach Fenstergrösse)
n_red_particles = 1
red_particles, = ax.plot([], [], marker = 'o', color = 'r', linestyle = '')   

# rect is the box edge
rect = plt.Rectangle(box.bounds[::2],
                     box.bounds[1] - box.bounds[0],
                     box.bounds[3] - box.bounds[2],
                     ec='none', lw=2, fc='none')
ax.add_patch(rect)
for key in ['right', 'bottom', 'top']:
    ax.spines[key].set_visible(False)
ax.set_xticks([])
ax.text (0,-0.05,"Toggeln: Pause: p, Bilder speichern: b (bedingt blit=False), Diagramm: S,s,z,Z")

    
ax2.set_ylim([0,1.55])
ax2.set_xlim([0,1])
ax2.set_xticks([])


# ----------------------------------------------------------------------------------
# Eventsteuerung

skala_neu = False
animation_paused = False
bilderspeichern = False


def tastendruck(event):
    global bilderspeichern, animation_paused
    print('you pressed: key, xdata, ydata', event.key, event.xdata, event.ydata)
    if event.key == "p":
        if animation_paused:
            ani.event_source.start()
        else:
            ani.event_source.stop()
        animation_paused = not animation_paused
    if event.key == "b":
        bilderspeichern = not bilderspeichern
    if event.key == "S":
        epot_boden.e1 = 0.6
        epot_boden.ea1 = 0.05
        epot_boden.e2 = 0.1
        box.Ekin_tot_0 = box.Ekin_tot
        box.konstant = "ekin"
        box.Q_mehrere_niveaus = []
        box.i_count = 0
        epot_boden.boden_zeichnen()  
    if event.key == "s":
        epot_boden.e1 = 0.2
        epot_boden.ea1 = 0.05
        epot_boden.e2 = 0.1
        box.Ekin_tot_0 = box.Ekin_tot
        box.konstant = "ekin"
        box.Q_mehrere_niveaus = []
        box.i_count = 0
        epot_boden.boden_zeichnen()  
    if event.key == "Z":
        epot_boden.e1 = 0.1
        epot_boden.ea1 = 0.55
        epot_boden.e2 = 0.6
        box.Ekin_tot_0 = box.Ekin_tot
        box.konstant = "ekin"
        box.Q_mehrere_niveaus = []
        box.i_count = 0
        epot_boden.boden_zeichnen()  
    if event.key == "z":
        epot_boden.e1 = 0.1
        epot_boden.ea1 = 0.15
        epot_boden.e2 = 0.2
        box.Ekin_tot_0 = box.Ekin_tot
        box.konstant = "ekin"
        box.Q_mehrere_niveaus = []
        box.i_count = 0
        epot_boden.boden_zeichnen()  
    
cid = fig.canvas.mpl_connect('key_press_event', tastendruck)


class Schaltflaechenfunktionen:
    
    def energien_anpassen(self, n_teilchen_alt):
        n_teilchen_neu = len(box.state)
        box.Ekin_tot_0 = box.Ekin_tot_0 * (n_teilchen_neu)/n_teilchen_alt
        box.Epot_tot_0 = box.Epot_tot_0 * (n_teilchen_neu)/n_teilchen_alt
        box.Etot_tot_0 = box.Etot_tot_0 * (n_teilchen_neu)/n_teilchen_alt
        txtBx_ekin.set_val(round(box.Ekin_tot_0*10000)/10000)
        
    def plus_edukt(self,event):
        n_teilchen = 5
        n_teilchen_alt = len(box.state)
        init_state = -0.5 + np.random.random((n_teilchen, 4))
        init_state[:, 1] += 0.5
        init_state[:, 1] *= 1.0            #Startpositioinen in y-Richtung
        init_state[:, 0] += 0.5    
        init_state[:, 0] *= epot_boden.x_uz         #Startpositionen in x-Richtung. Man könnte die Teilchen auch über den ganzen Bereich verteilen oder auf die Produktseite verschieben
        init_state[:, 2:] *= 1             #Startgeschwindigkeiten justieren
            
        box.state = np.concatenate((box.state, init_state))
        box.M = box.masse * np.ones(box.state.shape[0])
        box.alle_anheben()

        self.energien_anpassen(n_teilchen_alt = n_teilchen_alt)


    def plus_produkt(self,event):
        n_teilchen = 5
        n_teilchen_alt = len(box.state)
        init_state = -0.5 + np.random.random((n_teilchen, 4))
        init_state[:, 1] += 0.5             #Startpositioinen in y-Richtung#Startpositioinen in y-Richtung
        init_state[:, 1] *= 1.0            
        init_state[:, 0] += 0.5             #Startpositionen in x-Richtung. Man könnte die Teilchen auch über den ganzen Bereich verteilen oder auf die Produktseite verschieben 
        init_state[:, 0] *= (1-epot_boden.x_uz)
        init_state[:, 0] += epot_boden.x_uz
        init_state[:, 2:] *= 1              #Startgeschwindigkeiten justieren
               
        box.state = np.concatenate((box.state, init_state))
        box.M = box.masse * np.ones(box.state.shape[0])
        box.alle_anheben()

        self.energien_anpassen(n_teilchen_alt = n_teilchen_alt)

    def minus_edukt(self,event):
        n_teilchen = 5
        n_teilchen_alt = len(box.state)
        indexe = np.where(box.state[:,0] < epot_boden.x_uz)
        indexe = indexe[0][:5]
        n_teilchen_weg = len(indexe)

        box.state = np.delete(box.state, indexe, axis=0)
        box.M = box.masse * np.ones(box.state.shape[0])
        
        self.energien_anpassen(n_teilchen_alt = n_teilchen_alt)


    def minus_produkt(self,event):
        n_teilchen = 5
        n_teilchen_alt = len(box.state)
        indexe = np.where(box.state[:,0] > epot_boden.x_uz)
        indexe = indexe[0][:5]
        n_teilchen_weg = len(indexe)
        box.state = np.delete(box.state, indexe, axis=0)
        box.M = box.masse * np.ones(box.state.shape[0])

        self.energien_anpassen(n_teilchen_alt = n_teilchen_alt)
        
    def beschleunigung(self,event, faktor):
        box.zwischenschritte = faktor
    
    def reset(self,event):
        global skala_neu, text_Q, blitten
        box.Q = []                         # damit die Variable nie ganz leer ist
        box.Ekin_sum = 0
        box.Ekin_zeitschritte = 0
        text_Q = [ax_Q_avg.text(0.2,0.5," ", verticalalignment = "center")]

        """
        box.Q_mittel_laenge = 1
        box.i_Q_mittel = 0
        box.Q_mittel_vals = []
        box.i_count = 0
        """
        box.Q_mittel_laenge = 1
        box.Q_initialisieren()
        ax_Q_avg.cla()
        ax_Q_avg.axis("off")
        ax2.set_ylim([0,1.55])
        if blitten == True:
            skala_neu = True



        #bei der Umstellung von Niveaus werden auf konstante Ekin umgestellt, weil sonst die Temperatur sich verändert
    def e1(self,expression):
        epot_boden.e1 = float(expression)
        box.Ekin_tot_0 = box.Ekin_tot
        box.konstant = "ekin"
        box.i_count = 0
        epot_boden.boden_zeichnen()
        #fig.canvas.resize_event()  #danach werden zahlen überschrieben

    def ea(self,expression):
        epot_boden.ea1 = float(expression)
        box.Q_mehrere_niveaus = []
        box.i_count = 0
        epot_boden.boden_zeichnen()
        
    def e2(self,expression):
        epot_boden.e2 = float(expression)
        box.Ekin_tot_0 = box.Ekin_tot
        box.konstant = "ekin"
        epot_boden.ea1 = max(epot_boden.e2+epot_boden.ea2-epot_boden.e1, 0)
        box.Q_mehrere_niveaus = []
        box.i_count = 0
        epot_boden.boden_zeichnen()

    def ekin(self,expression):
        ekin_alt = box.Ekin_tot
        ekin_neu = max(0,float(expression))
        vfaktor = (ekin_neu/ekin_alt)**0.5
        box.state[:,2:] *= vfaktor
        box.Etot_tot_0 = box.Etot_tot_0 - ekin_alt  + ekin_neu
        box.Ekin_tot_0 = ekin_neu
        box.konstant = "ekin"
 
    def redist_v(self,event):
        v_neu = -0.5 + np.random.random((len(box.state), 2))
        ekin_neu = np.sum(0.5 * box.M * np.sum(v_neu**2, axis=1))
        f = (box.Ekin_tot/ekin_neu)**0.5
        box.state[:,2:] = v_neu * f
        
    def xpos_uz(self,expression):
        epot_boden.x_uz = float(expression)
        epot_boden.boden_zeichnen()
        #box.Q_initialisieren()
        #box.i_count = 0
        #self.reset(expression)

    def cla(self,expression):
        global text_Q
        ax_Q_avg.cla()
        ax_Q_avg.axis("off")
        text_Q = [ax_Q_avg.text(0.2,0.5," ", verticalalignment = "center")]
        ax2.cla()
        ax2.set_xticks([])
        fig.canvas.resize_event()
        ax2.set_xlim([0,1])
        self.reset(expression)
        
    def fix_toggle(self,expression):
        if box.konstant == "etot":
            box.konstant = "ekin"
        else: 
            box.konstant = "etot"
            

callback = Schaltflaechenfunktionen()


ax_x1 = fig.add_subplot(gs[7, 8])
ax_x2 = fig.add_subplot(gs[7, 9])
ax_x10 = fig.add_subplot(gs[7, 10])
ax_x100 = fig.add_subplot(gs[7, 11])

btn_x1 = Button(ax_x1, 'x1')
btn_x2 = Button(ax_x2, 'x2')
btn_x10 = Button(ax_x10, 'x10')
btn_x100 = Button(ax_x100, 'x100')

btn_x1.on_clicked(lambda x: callback.beschleunigung(x,1))
btn_x2.on_clicked(lambda x: callback.beschleunigung(x,2))
btn_x10.on_clicked(lambda x: callback.beschleunigung(x,10))
btn_x100.on_clicked(lambda x: callback.beschleunigung(x,100))

ax_plus_edukt =  fig.add_subplot(gs[7, 12])
btn_plus_edukt = Button(ax_plus_edukt, '+5 Ed')
btn_plus_edukt.on_clicked(callback.plus_edukt)

ax_plus_produkt =  fig.add_subplot(gs[7, 13])
btn_plus_produkt = Button(ax_plus_produkt, '+5 Prd')
btn_plus_produkt.on_clicked(callback.plus_produkt)

ax_minus_edukt =  fig.add_subplot(gs[7, 14])
btn_minus_edukt = Button(ax_minus_edukt, '-5 Ed')
btn_minus_edukt.on_clicked(callback.minus_edukt)

ax_minus_produkt =  fig.add_subplot(gs[7, 15])
btn_minus_produkt = Button(ax_minus_produkt, '-5 Prd')
btn_minus_produkt.on_clicked(callback.minus_produkt)

ax_e1 = fig.add_subplot(gs[8, 11])
txtBx_e1 = TextBox(ax_e1, r"$E_1$")
txtBx_e1.on_submit(callback.e1)

ax_ea = fig.add_subplot(gs[8, 13])
txtBx_ea = TextBox(ax_ea, r"$E_A$")
txtBx_ea.on_submit(callback.ea)

ax_e2 = fig.add_subplot(gs[8, 15])
txtBx_e2 = TextBox(ax_e2, r"$E_2$")
txtBx_e2.on_submit(callback.e2)

ax_xpos_uz = fig.add_subplot(gs[8, 9])
txtBx_xpos_uz = TextBox(ax_xpos_uz, "x(ÜZ)\n<1")
txtBx_xpos_uz.on_submit(callback.xpos_uz)


ax_ekin = fig.add_subplot(gs[9, 9])
txtBx_ekin = TextBox(ax_ekin, r"$E_{kin}$")
txtBx_ekin.set_val(round(box.Ekin_tot_0*10000)/10000)   #unbedingt vor on_submit, weil hier wird submittet
txtBx_ekin.on_submit(callback.ekin)

ax_redist_v =  fig.add_subplot(gs[9, 10])                   # Geschwindigkeit neu verteilen, wenn Teilchen sich seitlich kaum mehr bewegen
btn_redist_v = Button(ax_redist_v, 'kick')
btn_redist_v.on_clicked(callback.redist_v)

ax_reset =  fig.add_subplot(gs[9, 11])
btn_reset = Button(ax_reset, 'reset')
btn_reset.on_clicked(callback.reset)

ax_cla =  fig.add_subplot(gs[9, 12])
btn_cla = Button(ax_cla, 'cla()')
btn_cla.on_clicked(callback.cla)

# mit diesem Knopf stellt man um von fixem Ekin auf fixes Etot und umgekehrt. 
ax_tgl =  fig.add_subplot(gs[9, 13])
btn_tgl = Button(ax_tgl, r"$E_{fix}$"+"\ntoggle")
btn_tgl.label.set_fontsize(8)
btn_tgl.on_clicked(callback.fix_toggle)

# ------------
# Anzeige des Durchschnitts

ax_Q_avg = fig.add_subplot(gs[9, 14:])
ax_Q_avg.axis("off")







# -----------------------------------------------------------------------------------
# Animation
textelement = ax.text(0.1,0.9,"",color="blue",size = 14)
def init():
    """initialize animation"""
    global box, rect, skala_txt, legend_step, text_Q
    particles.set_data([], [])
    red_particles.set_data([], [])
    rect.set_edgecolor('none')
    skala_txt = []
    ylimmax = 1.55
    
    text_Q = [ax_Q_avg.text(0.2,0.5," ", verticalalignment = "center")]  
    return [particles, red_particles, rect, ax, ax2, ax_Q_avg, epot_boden.bodenlinie, box.legende] + box.verlaeufe + [textelement]

def animate(i_anim):
    """perform animation step"""
    global box, rect, dt, ax, ax2, fig, bilderspeichern, i_bild, markersize, skala_txt, legend_step, text_Q, skala_neu, blitten
    for z in range(box.zwischenschritte + 1):
        box.step(dt)
    
    if i_anim == 100:
        if box.konstant=='ekin':
            if Ekin_nachjustieren == True: 
                # !!!!!!!!!!!!! hier wird Ekin nachträglich auf den aktuellen Wert gesetzt, da es zu Beginn viel kleiner ist.
                box.Ekin_tot_0 = box.Ekin_tot
                txtBx_ekin.set_val(round(box.Ekin_tot_0*10000)/10000) 
                print("Ekin_tot_0 neu korrigiert nach x Schritten", box.Ekin_tot_0)
        
    if i_anim % 500 == 0:
        print("i_anim {}, E_tot start {:.04g}, E_tot {:.04g}, aktuelle Ekin_tot {:.04g}, Epot_tot {:.04g}".format(i_anim, box.Etot_tot_0, box.Etot_tot, box.Ekin_tot, box.Epot_tot))
        deltaEpot_boden = (epot_boden.e2-epot_boden.e1)*box.masse *  box.G
        Ekin_mittel = box.Ekin_sum/box.n_teilchen/box.Ekin_zeitschritte
        print("delta Epot der Niveaus = {:.05g}, Ekin Zeitmittel pro Teilchen = {:.05g}, fixiert: {}".format(deltaEpot_boden, Ekin_mittel, box.konstant) )



    textelement = ax.text(0.1,0.9,"",color="blue",size = 14)
    
    # programmierte Veränderungen
    """
    if i_anim == 1000:
        textelement = ax.text(0.1,0.9,"Erwärmung setzt ein",color="red",size = 14, zorder=0)
        box.Ekin_tot_0 = 6
    elif i_anim == 2000:
        x0p = 0.01
        y0p = 0.8
        dxp = 0.9
        dyp = 0.19
        pg = [[x0p,y0p],[x0p+dxp,y0p],[x0p+dxp,y0p+dyp],[x0p,y0p+dyp]]
        print(pg)
        farbe = "w"
        decke = patches.Polygon(pg, ec=farbe,fc=farbe, zorder=1)
        textelement=ax.add_patch(decke)

    """

    
    # update pieces of the animation
    rect.set_edgecolor('k')
    particles.set_data(box.state[:, 0], box.state[:, 1])
    particles.set_markersize(box.markersize)
    red_particles.set_data(box.state[:n_red_particles, 0], box.state[:n_red_particles, 1])
    red_particles.set_markersize(box.markersize*1.01)    
    

    if len(box.Q[0])>0:
        Qarray = np.array(box.Q)
        Qmax = np.max(Qarray)
        for i in range(len(box.Q)):
            box.verlaeufe[i].set_data(box.t_Q[:len(box.Q[0])], box.Q[i])

            if i_anim % 10 == 0:
                Q_text = "<Q{}> = {:.04g} \nE_kin = {:.04g}".format(i,np.average(box.Q[i]), box.Ekin_tot)
                text_Q[0].set_text(Q_text)
        try:
            if ax2.get_ylim()[1] < Qmax * 1.2 or ax2.get_ylim()[1] > Qmax * 8 or skala_neu == True:
                if blitten == True:
                    ax2.cla()
                ax2.set_xticks([])
                ylimmax = np.max(box.Q)*1.5
                ax2.set_ylim([0,ylimmax])
                skala_neu = False 
        except Exception as e:
                print(e)      
    epot_boden.bodenlinie.set_data(epot_boden.x,epot_boden.y)
    returnlist = [particles, red_particles, rect, ax, ax2, ax_Q_avg, epot_boden.bodenlinie, box.legende] + box.verlaeufe  + [textelement]
    
    if bilderspeichern:
        #gleichzeitige Anzeige geht nur, wenn blit = False
        plt.savefig("animation_{:04d}.png".format(i_bild))
        i_bild += 1

        
    return returnlist

#ax2.legend()


try:
    ani = animation.FuncAnimation(fig, animate, frames=6000,
                                  interval=10, blit=blitten, init_func=init, repeat=True)
except Exception as e:
                print(e)

# save the animation as an mp4.  This requires ffmpeg or mencoder to be
# installed.  The extra_args ensure that the x264 codec is used, so that
# the video can be embedded in html5.  You may need to adjust this for
# your system: for more information, see
# http://matplotlib.sourceforge.net/api/animation_api.html

#ani.save('particle_box.mp4', fps=30, extra_args=['-vcodec', 'libx264'])


plt.show()